import numpy as np
import matplotlib.pyplot as plt
from QuantumTomography import Tomography
import QuantumTomography.TomoFunctions as qf
import os
import glob
import scipy.integrate as si
import qutip as qt


"""Class for Tomography"""


class QSTAnalysis:
    def __init__(self):
        """
        This class performs the State Tomography on the recorded data. The actual Tomography class is taken from
        http://research.physics.illinois.edu/QI/Photonics/tomography/
        This code is written by the Kwiat Quantum Information group from the University of Illinois.
        """
        self.qt = Tomography()      # Writes tomography class on variable

        self.delimiter = '\t'               # Delimiter in files.
        self.folder = None                  # Directory of the files.
        self.singles_name = None            # File name of the recorded singles.
        self.coincidences_name = None       # File name of the recorded coincidences.
        self.efficiencies_name = None       # File name of the measured efficiencies.

        self.window = np.array([])          # In nanoseconds.
        self.int_time = np.array([])        # In seconds.
        self.efficiencies = np.array([])    # Array for different detector efficiencies.
        self.singles = np.array([])         # Array for recorded singles.
        self.coincidences = np.array([])    # Array for recorded coincidences.
        self.measurements = np.array([])    # Array for measurement specification.

        self.rho = np.array([])             # Array for density matrix.
        self.intensity = None               # The predicted overall intensity used to normalize the state.
        self.fvalp = None           # Final value of the internal optimization function. Values greater than the number
                                    # of measurements indicate poor agreement with a quantum state.

        self.trace_rho = None               # Variable for trace of rho
        self.purity_rho = None              # Variable for purity of rho
        self.concurrence_rho = None         # Variable for concurrence of rho

        self.delta_t = None                 # Variable for time-bin spacing
        self.fiber = None                   # Variable for type of long-distance fiber

    @staticmethod
    def find_between_strings(s: str, start: str, end: str):
        """
        This function returns a string which is between two sub-strings of a string
        :param s: string to search in
        :param start: start substring
        :param end: end substring
        :return: string between start and end
        """
        return (s.split(start))[1].split(end)[0]

    def glob_files(self, path: str, verbose=False):
        """
        This function uses the glob package to collect all files of a folder with a certain file name
        and returns them sorted alphabetically.
        :param path: directory path to look in for files
        :param verbose: If true, print array of files.
        :return:
        """
        files = glob.glob(path, recursive=True)  # Gets a string array of all file paths matching conditions above
        files = np.transpose(files)         # Makes the transpose
        files = np.sort(files)              # Sort file names alphabetically
        # arrays for sorting purposes
        two_digits = np.array([])           # array for two digit v-stage positions
        two_digits_index = np.array([])     # array for the indices of the two digit elements

        for i in range(len(files)):         # Loop to look for two digit vpath positions
            s = self.find_between_strings(files[i], '_vpathvpos_', '_w')    # Returns sub-string from file name
            if len(s) < 3:                  # if sub-string is e.g. 87 (2-digits) and not 107 (3-digits)
                two_digits = np.append(two_digits, files[i])        # Store 2-digit filenames in array
                two_digits_index = np.append(two_digits_index, i)   # Store also the index
        files = np.delete(files, two_digits_index.astype(int))      # Deletes the two digit file paths
        files = np.hstack((np.sort(two_digits), files))             # Add 2-digit filenames again at beginning
        if verbose:                         # If True, print the file names
            print(files)
        return files                        # Returns the sorted files

    @staticmethod
    def save_concurrence_purity_dt_conerr_purerr(fiber_dir: str, dt, concurr, purity, con_err, pur_err, fname):
        """
        This function saves the extracted concurrence, purity, time-bin spacing lists including errors in a
        separate text file.
        :param fiber_dir: fiber directory
        :param dt: time-bin spacings list
        :param concurr: concurrence list
        :param purity: purity list
        :param con_err: concurrence errors
        :param pur_err: purity errors
        :param fname: filename for generated file.
        :return:
        """
        parent_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data')
        folder = os.path.join(parent_path, fiber_dir)
        data = np.transpose([dt, concurr, purity, con_err, pur_err])
        path = os.path.join(folder, fname)
        head = 'time-bin spacing [ps], concurrence, purity, concurrence_std, purity_std'
        np.savetxt(path, data, delimiter='\t', comments='#', header=head)

    def get_singles_and_coincidence_file(self, fiber_dir: str, verbose=False):
        """
        This function gets all the singles and coincidence file names of a certain directory.
        :param fiber_dir: fiber directory of the measurement
        :param verbose: If True, print feedback
        :return:
        """
        parent_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data')
        folder = os.path.join(parent_path, fiber_dir)
        s_path = os.path.join(folder, '*_Singles.txt')  # Sets the path to look for files
        c_path = os.path.join(folder, '*_Coincidences.txt')
        # ('**' stands for all further directories, '*.txt' stands for all files with ending .txt)
        if verbose:
            print('Coincidence files:')
        c_files = self.glob_files(path=c_path)  # All cc files sorted
        if verbose:
            print('Singles files:')
        s_files = self.glob_files(path=s_path)  # All s files sorted
        if verbose:
            if len(c_files) != len(s_files):
                print('WARNING!, Not equal number of singles and coincidence files!')
        return c_files, s_files

    def read_data_2qubit(self, window_used, int_time):
        """
        This function reads in the coincidence and single data and prepares the required array for accidental correction.
        :param window_used: Coincidence window in ns.
        :param int_time: Integration time of the measurement in s.
        :return:
        """
        # For accidental correction:
        # Make coincidence window array for accidental correction.
        self.window = np.empty(4)  # For 4 detectors.
        self.window.fill(window_used * 1e-9)
        # Make integration time array for accidental correction.
        self.int_time = np.empty(9)  # For 4 detectors.
        self.int_time.fill(int_time)
        # Read in singles data.
        dummy_s = np.genfromtxt(fname=self.singles_name, delimiter=self.delimiter, comments='#')
        self.singles = dummy_s[:, [2, 4, 3, 5]]  # Hs, Hi, Vs, Vi
        # For Tomography:
        # Read in coincidences data.
        dummy_c = np.genfromtxt(fname=self.coincidences_name, delimiter=self.delimiter, comments='#')
        self.coincidences = dummy_c[:, [2, 3, 4, 5]]  # Hs-Hi, Hs-Vi, Vs-Hi, Vs-Vi
        # Create measurement array.
        dummy_m = self.qt.getTomoInputTemplate(numBits=2, numDet=2)
        n_qubit = 2  # Number of qubits.
        self.measurements = dummy_m[:, 2 ** n_qubit + 2 * n_qubit + 1: 2 ** n_qubit + 4 * n_qubit + 1]

    def load_efficiencies(self, verbose=False):
        """
        This function loads the measured detection efficiencies for detection efficiency correction.
        :param verbose: If true, it prints the efficiencies.
        :return:
        """
        self.efficiencies = np.genfromtxt(fname=self.efficiencies_name, comments='#')
        if verbose:
            print(self.efficiencies)

    def run_2qubit_tomography(self, verbose=False):
        """
        This function performs the 2qubit tomography using the Maximum Likelihood Method.
        :param verbose: If True, print feedback
        :return:
        """
        # StateTomography runs by default the Maximum Likelihood Estimation method.
        dummy = self.qt.StateTomography(measurements=self.measurements, counts=self.coincidences,
                                        efficiency=self.efficiencies, time=self.int_time,
                                        singles=self.singles, window=self.window, method='MLE')
        self.rho = dummy[0]
        self.intensity = dummy[1]
        self.fvalp = dummy[2]
        if verbose:
            print('The estimated density matrix is: \n')
            print(self.rho)
            print('\n The estimated intensity for normalizing the state is : ' + str(self.intensity))
            print('The final value of the optimization function is: ' + str(self.fvalp))

    def save_density_matrix(self, multiple=False):
        """
        This function saves the reconstructed density matrix.
        :return:
        """
        #if multiple:
        rho_file = self.coincidences_name + '_RHO.txt'
        #else:
        #   path = os.path.join(pathlib.Path.home(), self.folder)
         #   rho_file = os.path.join(path, 'rho_' + self.coincidences_name)
        comment = 'HH,  HV, VH, VV'
        np.savetxt(rho_file, self.rho, delimiter=self.delimiter, header=comment, comments='#', fmt='%s')

    def read_density_matrix(self, rho_path):
        """
        This function is required for plotting an already constructed density matrix.
        :param rho_file:
        :return:
        """
        self.rho = np.genfromtxt(fname=rho_path, delimiter=self.delimiter, comments='#', dtype=str)
        self.convert_rho()
        self.extract_complex_numbers()
        self.analyse_rho()

    def load_density_matrix(self, rho_file):
        """
        Load measured density matrix
        :param class_object:
        :param rho_file:
        :return:
        """
        self.read_density_matrix(rho_file=rho_file)
        self.extract_complex_numbers()
        self.analyse_rho()

    def extract_complex_numbers(self, verbose=False):
        """
        This function extracts the real and complex part out of the (string) density matrix array.
        :param verbose: If True, return feedback
        :return:
        """
        # self.convert_rho()
        self.rho_real = np.zeros_like(self.rho)                 # Zero-array as shape rho for real part
        self.rho_imag = self.rho_real.copy()                    # Zero-array as shape rho for imaginary part
        for i in range(len(self.rho[:, 0])):
            for j in range(len(self.rho[0, :])):
                self.rho_real[i, j] = complex(self.rho[i, j]).real      # Extract real part
                self.rho_imag[i, j] = complex(self.rho[i, j]).imag      # Extract imaginary part
        self.rho_real = self.rho_real.astype(float)                     # Convert strings to float
        self.rho_imag = self.rho_imag.astype(float)                     # Convert strings to float
        if verbose:                                 # If True, give feedback
            print('Real part of rho:')
            print(self.rho_real)
            print('\n')
            print('Imaginary part of rho:')
            print(self.rho_imag)
            print('\n')
        self.rho = self.rho_real + 1j * self.rho_imag

    def convert_rho(self):
        """
        This function deals with the array structure returned by Kwiat's tomography package
        :return:
        """
        for i in range(len(self.rho[:, 0])):
            for j in range(len(self.rho[0, :])):
                self.rho[i, j] = self.rho[i, j].replace(' ', '')
                self.rho[i, j] = self.rho[i, j].replace('(', '')
                self.rho[i, j] = self.rho[i, j].replace(')', '')
                self.rho[i, j] = complex(self.rho[i, j])

    def get_trace(self, verbose=False):
        """
        Extracts trace of rho.
        :param verbose: If true, print feedback.
        :return:
        """
        self.trace_rho = np.real(np.trace(self.rho))
        if verbose:
            print(r'$\Tr{\rho} = $' + str(self.trace_rho))

    def get_purity(self, verbose=False):
        """
        Extracts purity of rho.
        :param verbose: If true, print feedback.
        :return:
        """
        self.purity_rho = qf.purity(self.rho)
        if verbose:
            print(r'$\Tr{{\rho} ** 2} = $#' + str(self.purity_rho))

    def get_concurrence(self, verbose=False):
        """
        Extracts concurrence of rho.
        :param verbose: If true, print feedback.
        :return:
        """
        self.concurrence_rho = qf.concurrence(self.rho)
        if verbose:
            print(r'$C({\rho}) = $' + str(self.concurrence_rho))

    def analyse_rho(self):
        """
        Extracts trace, purity and concurrence of rho.
        :param two_qubit: if True, get concurrence.
        :return:
        """
        self.get_trace()
        self.get_purity()
        self.get_concurrence()


"""Class for Monte Carlo simulation"""


class MonteCarlo:
    def __init__(self):
        pass

    @staticmethod
    def get_probs_from_counts(measured_counts):
        """
        This function calculates from a given count array the corresponding probabilities. It returns an array with
        probabilities of same shape as the input count array and a 1d-array if total counts per line (bases setting) in the
        input count array
        :param measured_counts: Experimentally measured counts as array. Each line is a specific basis setting
        :return:
        """
        prob_arr = np.zeros_like(measured_counts)
        tot_counts = np.zeros(len(measured_counts))

        for base_line in range(len(measured_counts)):
            n = sum(measured_counts[base_line, :])
            prob_arr[base_line, :] = measured_counts[base_line, :] / n
            tot_counts[base_line] = n
        return prob_arr, tot_counts

    @staticmethod
    def monte_carlo_step(total_count_vector, probability_array, verbose=False):
        """
        This function evaluates a single run in the MC-simulation. It distributes the total recorded counts according to
        the given probabilities in a multinomial distribution.
        :param total_count_vector:
        :param probability_array:
        :param verbose:
        :return:
        """
        mc_counts = np.zeros_like(probability_array)
        for base_line in range(len(total_count_vector)):
            mc_counts[base_line, :] = np.random.multinomial(total_count_vector[base_line], probability_array[base_line, :])
        if verbose:
            print(mc_counts)
        return mc_counts


"""Function to generate 3D bar plots"""


def barplot_3d(matrix, z_label: str, figure, azim_view=-63.5, elev_view=23., color='jet',
               sub_row=1, sub_col=1, sub_idx=1, ticksize=18, xtick_rotation=0, ytick_rotation=0):
    """
    This function generates a 3D bar plot of a 2-qubit matrix
    :param matrix: input 4x4 matrix
    :param z_label: specify z-axis label
    :param figure: input figure variable
    :param azim_view: azimuthal viewing angle
    :param elev_view: elevation viewing angle
    :param color: Specific colormap
    :param sub_row: Rows of generated subplot
    :param sub_col: Columns of generated subplot
    :param sub_idx: index of subplot
    :param ticksize: text size of the xyz-ticks
    :param xtick_rotation: rotation angle of x-ticks
    :param ytick_rotation: rotation angle of y-ticks
    :return:
    """
    # Process input matrix for 3D-bar plot
    x = np.arange(1, len(matrix[:, 0]) + 1, 1, dtype=float)
    y = np.copy(x)
    xpos, ypos = np.meshgrid(x, y)
    xpos = xpos.flatten()
    ypos = ypos.flatten()
    zpos = np.zeros_like(xpos)
    dx = 1 * np.ones_like(zpos) # adjust bar width here, maybe 0.9
    dy = dx.copy()
    dz = matrix.flatten()

    colorMap = plt.cm.get_cmap(color)       # generate a colormap
    # Generate 3D-bar plot
    ax = figure.add_subplot(sub_row, sub_col, sub_idx, projection='3d')
    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, color=colorMap((dz + 1) / 2))
    tick = np.array(['HH', 'HV', 'VH', 'VV'])               # Specify the axis ticks
    ticks_loc = np.arange(1.5, 5.5, 1)                      # Specify tick position

    plt.xticks(ticks_loc, tick, rotation=xtick_rotation)    # Set x-axis ticks and rotation
    plt.yticks(ticks_loc, tick, rotation=ytick_rotation)    # # Set y-axis ticks and rotation
    ax.set_zticks([-1, -0.5, 0, 0.5, 1])                    # Set z-ticks
    ax.azim = azim_view                                     # Set azimuthal viewing angle
    ax.elev = elev_view                                     # Set elevation viewing angle
    ax.dist = 8                                             # Set viewing distance
    ax.set_zlim3d(-1, 1)                                    # Set z-axis limit
    ax.zaxis.set_rotate_label(False)                        # No rotation in z-axis label
    ax.set_zlabel(z_label, fontsize=20, labelpad=15)        # Set z-label

    plt.xticks(fontsize=ticksize)                           # Plot x-ticks
    plt.yticks(fontsize=ticksize)                           # Plot y-ticks
    for t in ax.zaxis.get_major_ticks():                    # Set fontsize for z-ticks
        t.label.set_fontsize(18)
    return ax                                               # Returns axis variable for further editing


def label_4t4_matrix(one: str, two: str, three: str, four: str):
    """
    This function adds tick labels to a 4x4 matrix
    :param one:
    :param two:
    :param three:
    :param four:
    :return:
    """
    tick = np.array([one, two, three, four])
    ticks_loc = np.arange(1.5, 5.5, 1)
    plt.xticks(ticks_loc, tick)
    plt.yticks(ticks_loc, tick)


"""Functions to handle data imports and plotting preparations"""


def load_columns_from_file(path: str, col_num: int, delimiter='\t', no_of_headers=0):
    """
    This function loads one to five column data. Note, it can be easly adjusted for more.
    :param path: path of data file.
    :param col_num: number of columns in file.
    :param delimiter: Delimiter in text file.
    :param no_of_headers: Number of header lines in file to skip (comments marked with # are skiped automatically)
    :return:
    """
    dummy = np.genfromtxt(path, comments='#', delimiter=delimiter, skip_header=no_of_headers)
    if col_num >= 1:
        x = dummy[:, 0]
        if col_num == 1:
            return x
    if col_num >= 2:
        y = dummy[:, 1]
        if col_num == 2:
            return x, y
    if col_num >= 3:
        z = dummy[:, 2]
        if col_num == 3:
            return x, y, z
    if col_num >= 4:
        a = dummy[:, 3]
        if col_num == 4:
            return x, y, z, a
    if col_num >= 5:
        b = dummy[:, 4]
        if col_num == 5:
            return x, y, z, a, b


def find_between_strings(s, start, end):
    """
    This function returns a string which is between two sub-strings of a string
    :param s: string to search in
    :param start: start substring
    :param end: end substring
    :return: string between start and end
    """
    return (s.split(start))[1].split(end)[0]


def generate_centered_x_axis(array_x, array_y):
    """
    Generates an array of same length as input array with the maximum of array y at 0.
    To ensure same length, the first element of the new array is deleted (acceptable error for visualizing)
    :param array_x: x-axis data
    :param array_y: y-axis data
    :return:
    """
    center_idx = np.argmax(array_y)
    no_pos_elements = len(array_x) - center_idx
    no_neg_elements = center_idx
    a_pos = np.arange(0, no_pos_elements, 1)
    a_neg = np.arange(-1 * no_neg_elements, 0, 1)
    a = np.hstack((a_neg, a_pos))
    return a


def find_nearest(array, value):
    """
    This function finds the nearest element in an array to a specific value
    :param array:
    :param value:
    :return:
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx


def matlab_to_python_complex(matrix_matlab):
    """
    Transfers complex matlab syntax to python syntax.
    :param matrix_matlab:
    :return:
    """
    # Replace complex unit and eliminate spaces
    for i in range(len(matrix_matlab[:, 0])):
        for j in range(len(matrix_matlab[0, :])):
            matrix_matlab[i, j] = matrix_matlab[i, j].replace('i', 'j')
            matrix_matlab[i, j] = matrix_matlab[i, j].replace(' ', '')
    # Create matrix of same shape
    matrix_python = np.zeros_like(matrix_matlab, dtype=complex)
    # Transform strings from dummy_chi to complex numbers in chi
    for i in range(len(matrix_matlab[:, 0])):
        for j in range(len(matrix_matlab[0, :])):
            matrix_python[i, j] = matrix_matlab[i, j].astype(complex)
    return matrix_python


def load_chi(file_path):
    """
    This function loads in the chi_matrix
    :param file_path: Path of the chi-matrix
    :return:
    """
    dummy_chi = np.genfromtxt(fname=file_path,
                              delimiter='\t', comments='#', dtype=str)
    chi = matlab_to_python_complex(matrix_matlab=dummy_chi)
    return chi


"""Functions for simulating figure 4"""


def get_errorcount_simulation(init_rho, window_no, fiber):
    """
    This function returns the simulation for concurrence and purity
    :param init_rho: Initial density matrix
    :param window_no: 1, 2, 3 is multiplied with the sigma to get the window
    :param fiber: fiber description ['hcf', 'smf']
    :return: time_spacing, purity, concurrence
    """
    init_rho = qt.Qobj(init_rho, dims=[[2, 2], [2, 2]])
    assert fiber in ['hcf', 'smf'], 'Fiber type is not available!'
    if fiber == 'hcf':
        sigma = 23.111
    elif fiber == 'smf':
        sigma = 53.686
    window = window_no * sigma
    purity_arr = np.array([])
    concurr_arr = np.array([])
    spacings = np.arange(0, 521, 2)
    for dt in spacings:
        eps_0 = si.quad(gauss, a=-window, b=window, args=(0, sigma))    # Center
        eps_0 = eps_0[0]

        hh_side_amplitude = 0.5
        vv_side_amplitude = 1 - hh_side_amplitude
        eps_hh = si.quad(gauss, a=-window, b=window, args=(dt, sigma, hh_side_amplitude))  # H-pulse
        eps_hh = eps_hh[0]
        eps_vv = si.quad(gauss, a=-window, b=window, args=(dt, sigma, vv_side_amplitude))   # V-pulse
        eps_vv = eps_vv[0]
        rho_out = overlap_errorcount(rho_in=init_rho, eps_0=eps_0, eps_hh=eps_hh, eps_vv=eps_vv)
        purity_arr = np.append(purity_arr, rho_out.purity())
        concurr_arr = np.append(concurr_arr, qt.concurrence(rho_out))
    return spacings, purity_arr, concurr_arr


def gauss(x, x0, sigma, amp=1):
    """
    Gaus pulse shape
    :param x: abscissa
    :param x0: center of gaus
    :param sigma: standard deviation
    :param amp: Amplitude of the gauss
    :return:
    """
    g = amp * (1 / (np.sqrt(2 * np.pi) * sigma)) * np.e ** (-1 * ((x - x0) ** 2) / (2 * sigma ** 2))
    return g


def overlap_errorcount(rho_in, eps_0, eps_hh, eps_vv):
    hh, hv, vh, vv = form_2qubit_basis()
    hh_mixing_rho = hh * hh.dag()
    vv_mixing_rho = vv * vv.dag()
    rho_out = (eps_0 / (eps_0 + eps_hh + eps_vv)) * rho_in + (eps_hh / (eps_0 + eps_hh + eps_vv)) * hh_mixing_rho\
            + (eps_vv / (eps_0 + eps_hh + eps_vv)) * vv_mixing_rho
    return rho_out / rho_out.tr()


def form_2qubit_basis():
    """
    Define 2-qubit polarization basis.
    :return:
    """
    h = qt.basis(2, 0)
    v = qt.basis(2, 1)

    hh = qt.composite(h, h)
    hv = qt.composite(h, v)
    vh = qt.composite(v, h)
    vv = qt.composite(v, v)
    return hh, hv, vh, vv
